@@ -1,21 +1,44 @@
 # -*- coding: utf-8 -*-
 from __future__ import unicode_literals
 
+from os import environ
+
 from ..concurrency import WorkerPool
 from ..lock import softlock_add, softlock_list, softlock_remove
 from ..utils.cmdline import get_target_nodes
-from ..utils.text import blue, bold, cyan, error_summary, green, mark_for_translation as _, \
-    randstr
+from ..utils.table import ROW_SEPARATOR, render_table
+from ..utils.text import bold, error_summary, green, mark_for_translation as _, randstr, red, \
+    yellow
 from ..utils.time import format_timestamp
-from ..utils.ui import io
+from ..utils.ui import io, page_lines
+
+
+def remove_dummy_nodes(targets):
+    _targets = []
+    for node in targets:
+        if node.dummy:
+            io.stdout(_("{x} {node}  is a dummy node").format(node=bold(node.name), x=yellow("»")))
+        else:
+            _targets.append(node)
+    return _targets
+
+
+def remove_lock_if_present(node, lock_id):
+    for lock in softlock_list(node):
+        if lock['id'] == lock_id:
+            softlock_remove(node, lock_id)
+            return True
+    return False
 
 
 def bw_lock_add(repo, args):
     errors = []
     target_nodes = get_target_nodes(repo, args['target'], adhoc_nodes=args['adhoc_nodes'])
+    target_nodes = remove_dummy_nodes(target_nodes)
     pending_nodes = target_nodes[:]
     max_node_name_length = max([len(node.name) for node in target_nodes])
     lock_id = randstr(length=4).upper()
+    io.progress_set_total(len(pending_nodes))
 
     def tasks_available():
         return bool(pending_nodes)
@@ -34,6 +57,7 @@ def bw_lock_add(repo, args):
         }
 
     def handle_result(task_id, return_value, duration):
+        io.progress_advance()
         io.stdout(_("{x} {node}  locked with ID {id} (expires in {exp})").format(
             x=green("✓"),
             node=bold(task_id.ljust(max_node_name_length)),
@@ -64,8 +88,10 @@ def bw_lock_add(repo, args):
 def bw_lock_remove(repo, args):
     errors = []
     target_nodes = get_target_nodes(repo, args['target'], adhoc_nodes=args['adhoc_nodes'])
+    target_nodes = remove_dummy_nodes(target_nodes)
     pending_nodes = target_nodes[:]
     max_node_name_length = max([len(node.name) for node in target_nodes])
+    io.progress_set_total(len(pending_nodes))
 
     def tasks_available():
         return bool(pending_nodes)
@@ -73,17 +99,27 @@ def bw_lock_remove(repo, args):
     def next_task():
         node = pending_nodes.pop()
         return {
-            'target': softlock_remove,
+            'target': remove_lock_if_present,
             'task_id': node.name,
             'args': (node, args['lock_id'].upper()),
         }
 
     def handle_result(task_id, return_value, duration):
-        io.stdout(_("{x} {node}  lock {id} removed").format(
-            x=green("✓"),
-            node=bold(task_id.ljust(max_node_name_length)),
-            id=args['lock_id'].upper(),
-        ))
+        io.progress_advance()
+        if return_value is True:
+            io.stdout(_("{x} {node}  lock {id} removed").format(
+                x=green("✓"),
+                node=bold(task_id.ljust(max_node_name_length)),
+                id=args['lock_id'].upper(),
+            ))
+        else:
+            io.stderr(_(
+                "{x} {node}  has no lock with ID {id}"
+            ).format(
+                x=red("!"),
+                node=bold(task_id.ljust(max_node_name_length)),
+                id=args['lock_id'].upper(),
+            ))
 
     def handle_exception(task_id, exception, traceback):
         msg = "{}: {}".format(task_id, exception)
@@ -108,8 +144,8 @@ def bw_lock_remove(repo, args):
 def bw_lock_show(repo, args):
     errors = []
     target_nodes = get_target_nodes(repo, args['target'], adhoc_nodes=args['adhoc_nodes'])
+    target_nodes = remove_dummy_nodes(target_nodes)
     pending_nodes = target_nodes[:]
-    max_node_name_length = max([len(node.name) for node in target_nodes])
     locks_on_node = {}
 
     def tasks_available():
@@ -125,6 +161,7 @@ def bw_lock_show(repo, args):
 
     def handle_result(task_id, return_value, duration):
         locks_on_node[task_id] = return_value
+        repo.hooks.lock_show(repo, repo.get_node(task_id), return_value)
 
     def handle_exception(task_id, exception, traceback):
         msg = "{}: {}".format(task_id, exception)
@@ -147,77 +184,50 @@ def bw_lock_show(repo, args):
         error_summary(errors)
         return
 
-    headers = (
-        ('id', _("ID")),
-        ('formatted_date', _("Created")),
-        ('formatted_expiry', _("Expires")),
-        ('user', _("User")),
-        ('items', _("Items")),
-        ('comment', _("Comment")),
-    )
+    rows = [[
+        bold(_("node")),
+        bold(_("ID")),
+        bold(_("created")),
+        bold(_("expires")),
+        bold(_("user")),
+        bold(_("items")),
+        bold(_("comment")),
+    ], ROW_SEPARATOR]
 
-    locked_nodes = 0
-    for node_name, locks in locks_on_node.items():
-        if locks:
-            locked_nodes += 1
-
-    previous_node_was_unlocked = False
-    for node_name, locks in sorted(locks_on_node.items()):
-        if not locks:
-            io.stdout(_("{x} {node}  no soft locks present").format(
-                x=green("✓"),
-                node=bold(node_name.ljust(max_node_name_length)),
-            ))
-            previous_node_was_unlocked = True
-
-    output_counter = 0
     for node_name, locks in sorted(locks_on_node.items()):
         if locks:
-            # Unlocked nodes are printed without empty lines in
-            # between them. Locked nodes can produce lengthy output,
-            # though, so we add empty lines.
-            if (
-                previous_node_was_unlocked or (
-                    output_counter > 0 and output_counter < locked_nodes
-                )
-            ):
-                previous_node_was_unlocked = False
-                io.stdout('')
-
+            first_lock = True
             for lock in locks:
                 lock['formatted_date'] = format_timestamp(lock['date'])
                 lock['formatted_expiry'] = format_timestamp(lock['expiry'])
+                first_item = True
+                for item in lock['items']:
+                    rows.append([
+                        node_name if first_item and first_lock else "",
+                        lock['id'] if first_item else "",
+                        lock['formatted_date'] if first_item else "",
+                        lock['formatted_expiry'] if first_item else "",
+                        lock['user'] if first_item else "",
+                        item,
+                        lock['comment'] if first_item else "",
+                    ])
+                    # always repeat for grep style
+                    first_item = environ.get("BW_TABLE_STYLE") == 'grep'
+                # always repeat for grep style
+                first_lock = environ.get("BW_TABLE_STYLE") == 'grep'
+        else:
+            rows.append([
+                node_name,
+                _("(none)"),
+                "",
+                "",
+                "",
+                "",
+                "",
+            ])
+        rows.append(ROW_SEPARATOR)
 
-            lengths = {}
-            headline = "{x} {node}  ".format(
-                x=blue("i"),
-                node=bold(node_name.ljust(max_node_name_length)),
-            )
-
-            for column, title in headers:
-                lengths[column] = len(title)
-                for lock in locks:
-                    if column == 'items':
-                        length = max([len(selector) for selector in lock[column]])
-                    else:
-                        length = len(lock[column])
-                    lengths[column] = max(lengths[column], length)
-                headline += bold(title.ljust(lengths[column] + 2))
-
-            io.stdout(headline.rstrip())
-            for lock in locks:
-                for lineno, item_selectors in enumerate(lock['items']):
-                    line = "{x} {node}  ".format(
-                        x=cyan("›"),
-                        node=bold(node_name.ljust(max_node_name_length)),
-                    )
-                    for column, title in headers:
-                        if column == 'items':
-                            line += lock[column][lineno].ljust(lengths[column] + 2)
-                        elif lineno == 0:
-                            line += lock[column].ljust(lengths[column] + 2)
-                        else:
-                            line += " " * (lengths[column] + 2)
-                    io.stdout(line.rstrip())
-
-            output_counter += 1
+    page_lines(render_table(
+        rows[:-1],  # remove trailing ROW_SEPARATOR
+        alignments={1: 'center'},
+    ))
