@@ -2,7 +2,7 @@
 from __future__ import unicode_literals
 
 from argparse import ArgumentParser
-from os import environ
+from os import environ, getcwd
 
 from .. import VERSION_STRING
 from ..utils.text import mark_for_translation as _
@@ -56,6 +56,15 @@ def build_parser_bw():
         dest='debug',
         help=_("print debugging info (implies -v)"),
     )
+    parser.add_argument(
+        "-r",
+        "--repo-path",
+        default=environ.get('BW_REPO_PATH', getcwd()),
+        dest='repo_path',
+        help=_("Look for repository at this path (defaults to current working directory)"),
+        metavar=_("DIRECTORY"),
+        type=str,
+    )
     parser.add_argument(
         "--version",
         action='version',
@@ -135,6 +144,25 @@ def build_parser_bw():
         metavar=_("SELECTOR"),
         type=str,
     )
+    parser_apply.add_argument(
+        "-S",
+        "--no-summary",
+        action='store_false',
+        dest='summary',
+        help=_("don't show stats summary"),
+    )
+    parser_apply.add_argument(
+        "-r",
+        "--resume-file",
+        default=None,
+        dest='resume_file',
+        help=_(
+            "path to a file that a list of completed nodes will be added to; "
+            "if the file already exists, any nodes therein will be skipped"
+        ),
+        metavar=_("PATH"),
+        type=str,
+    )
 
     # bw debug
     help_debug = _("Start an interactive Python shell for this repository")
@@ -226,6 +254,20 @@ def build_parser_bw():
         type=str,
         help=_("list items for this node"),
     )
+    parser_items.add_argument(
+        'item',
+        metavar=_("ITEM"),
+        nargs='?',
+        type=str,
+        help=_("show configuration for this item"),
+    )
+    parser_items.add_argument(
+        'attr',
+        metavar=_("ATTRIBUTE"),
+        nargs='?',
+        type=str,
+        help=_("show only this item attribute"),
+    )
     parser_items.add_argument(
         "-f",
         "--file-preview",
@@ -233,7 +275,7 @@ def build_parser_bw():
         help=_("print preview of given file"),
         metavar=_("FILE"),
         required=False,
-        type=str,
+        type=str,  # TODO 3.0 convert to bool and use ITEM arg
     )
     parser_items.add_argument(
         "-w",
@@ -251,6 +293,12 @@ def build_parser_bw():
         dest='show_repr',
         help=_("show more verbose representation of each item"),
     )
+    parser_items.add_argument(
+        "--state",
+        action='store_true',
+        dest='show_sdict',
+        help=_("show actual item status on node instead of should-be configuration"),
+    )
 
     # bw lock
     help_lock = _("Manage locks on nodes used to prevent collisions between BundleWrap users")
@@ -374,11 +422,30 @@ def build_parser_bw():
     )
     parser_metadata.set_defaults(func=bw_metadata)
     parser_metadata.add_argument(
-        'node',
+        'target',
         metavar=_("NODE"),
         type=str,
         help=_("node to print JSON-formatted metadata for"),
     )
+    parser_metadata.add_argument(
+        'keys',
+        default=[],
+        metavar=_("KEY"),
+        nargs='*',
+        type=str,
+        help=_("print only partial metadata from the given space-separated key path"),
+    )
+    parser_metadata.add_argument(
+        "-t",
+        "--table",
+        action='store_true',
+        dest='table',
+        help=_(
+            "show a table of selected metadata values from multiple nodes instead; "
+            "allows for multiple comma-separated paths in KEY; "
+            "allows for node selectors in NODE (e.g. 'NODE1,NODE2,GROUP1,bundle:BUNDLE1...')"
+        ),
+    )
 
     # bw nodes
     help_nodes = _("List all nodes in this repository")
@@ -653,6 +720,7 @@ def build_parser_bw():
     parser_run.add_argument(
         'command',
         metavar=_("COMMAND"),
+        nargs='+',
         type=str,
         help=_("command to run"),
     )
@@ -679,6 +747,18 @@ def build_parser_bw():
                "(defaults to {})").format(bw_run_p_default),
         type=int,
     )
+    parser_run.add_argument(
+        "-r",
+        "--resume-file",
+        default=None,
+        dest='resume_file',
+        help=_(
+            "path to a file that a list of completed nodes will be added to; "
+            "if the file already exists, any nodes therein will be skipped"
+        ),
+        metavar=_("PATH"),
+        type=str,
+    )
 
     # bw stats
     help_stats = _("Show some statistics about your repository")
