@@ -4,11 +4,13 @@ from __future__ import unicode_literals
 from copy import copy
 from sys import exit
 
+from ..exceptions import ItemDependencyLoop
 from ..concurrency import WorkerPool
 from ..plugins import PluginManager
 from ..repo import Repository
-from ..utils.cmdline import get_target_nodes
-from ..utils.text import bold, green, mark_for_translation as _, red
+from ..utils.cmdline import count_items, get_target_nodes
+from ..utils.plot import explain_item_dependency_loop
+from ..utils.text import bold, green, mark_for_translation as _, red, yellow
 from ..utils.ui import io
 
 
@@ -18,6 +20,21 @@ def bw_test(repo, args):
     else:
         pending_nodes = copy(list(repo.nodes))
 
+        # Print warnings for unused bundles. Only do this if we are to
+        # test the entire repo, though.
+        # TODO 3.0 Orphaned bundles should be errors (maybe optionally)
+        orphaned_bundles = set(repo.bundle_names)
+        for node in repo.nodes:
+            for bundle in node.bundles:
+                orphaned_bundles.discard(bundle.name)
+        for bundle in sorted(orphaned_bundles):
+            io.stdout(_("{x} {bundle}  is an unused bundle").format(
+                bundle=bold(bundle),
+                x=yellow("!"),
+            ))
+
+    io.progress_set_total(count_items(pending_nodes))
+
     def tasks_available():
         return bool(pending_nodes)
 
@@ -32,14 +49,23 @@ def bw_test(repo, args):
             },
         }
 
+    def handle_exception(task_id, exception, traceback):
+        if isinstance(exception, ItemDependencyLoop):
+            for line in explain_item_dependency_loop(exception, task_id):
+                io.stderr(line)
+        raise exception
+
     worker_pool = WorkerPool(
         tasks_available,
         next_task,
+        handle_exception=handle_exception,
         pool_id="test",
         workers=args['node_workers'],
     )
     worker_pool.run()
 
+    io.progress_set_total(0)
+
     checked_groups = []
     for group in repo.groups:
         if group in checked_groups:
