@@ -1,7 +1,7 @@
 # -*- coding: utf-8 -*-
 from __future__ import unicode_literals
 
-from .exceptions import BundleError, NoSuchItem
+from .exceptions import BundleError, ItemDependencyError, NoSuchItem
 from .items import Item
 from .items.actions import Action
 from .utils.text import mark_for_translation as _
@@ -106,13 +106,11 @@ def _find_items_of_types(item_types, items, include_dummy=False):
     """
     Returns a subset of items with any of the given types.
     """
-    return list(filter(
-        lambda item:
-            item.id.split(":", 1)[0] in item_types and (
-                include_dummy or not isinstance(item, DummyItem)
-            ),
-        items,
-    ))
+    for item_id, item in items.items():
+        if item_id.split(":", 1)[0] in item_types and (
+            include_dummy or not isinstance(item, DummyItem)
+        ):
+            yield item
 
 
 def _flatten_dependencies(items):
@@ -120,33 +118,42 @@ def _flatten_dependencies(items):
     This will cause all dependencies - direct AND inherited - to be
     listed in item._flattened_deps.
     """
-    for item in items:
-        item._flattened_deps = list(set(
-            item._deps + _get_deps_for_item(item, items)
-        ))
+    for item in items.values():
+        if not hasattr(item, '_flattened_deps'):
+            _flatten_deps_for_item(item, items)
     return items
 
 
-def _get_deps_for_item(item, items, deps_found=None):
+def _flatten_deps_for_item(item, items):
     """
     Recursively retrieves and returns a list of all inherited
     dependencies of the given item.
 
-    Note: This can handle loops, but won't detect them.
+    This can handle loops, but will ignore them.
     """
-    if deps_found is None:
-        deps_found = []
-    deps = []
+    item._flattened_deps = set(item._deps)
+
     for dep in item._deps:
-        if dep not in deps_found:
-            deps.append(dep)
-            deps_found.append(dep)
-            deps += _get_deps_for_item(
-                find_item(dep, items),
-                items,
-                deps_found,
-            )
-    return deps
+        try:
+            dep_item = items[dep]
+        except KeyError:
+            raise ItemDependencyError(_(
+                "'{item}' in bundle '{bundle}' has a dependency (needs) "
+                "on '{dep}', which doesn't exist"
+            ).format(
+                item=item.id,
+                bundle=item.bundle.name,
+                dep=dep,
+            ))
+        # Don't recurse if we have already resolved nested dependencies
+        # for this item. Also serves as a guard against infinite
+        # recursion when there are loops.
+        if not hasattr(dep_item, '_flattened_deps'):
+            _flatten_deps_for_item(dep_item, items)
+
+        item._flattened_deps |= set(dep_item._flattened_deps)
+
+    item._flattened_deps = sorted(item._flattened_deps)
 
 
 def _has_trigger_path(items, item, target_item_id):
@@ -173,13 +180,14 @@ def _inject_bundle_items(items):
     Adds virtual items that depend on every item in a bundle.
     """
     bundle_items = {}
-    for item in items:
+    for item in items.values():
         if item.bundle is None:
             continue
         if item.bundle.name not in bundle_items:
             bundle_items[item.bundle.name] = BundleItem(item.bundle)
         bundle_items[item.bundle.name]._deps.append(item.id)
-    return list(bundle_items.values()) + items
+    items.update({item.id: item for item in bundle_items.values()})
+    return items
 
 
 def _inject_canned_actions(items):
@@ -188,7 +196,7 @@ def _inject_canned_actions(items):
     triggers and adds them to the list of items.
     """
     added_actions = {}
-    for item in items:
+    for item in items.values():
         for triggered_item_id in item.triggers:
             if triggered_item_id in added_actions:
                 # action has already been triggered
@@ -203,8 +211,8 @@ def _inject_canned_actions(items):
             target_item_id = "{}:{}".format(type_name, item_name)
 
             try:
-                target_item = find_item(target_item_id, items)
-            except NoSuchItem:
+                target_item = items[target_item_id]
+            except KeyError:
                 raise BundleError(_(
                     "{item} in bundle '{bundle}' triggers unknown item '{target_item}'"
                 ).format(
@@ -236,32 +244,65 @@ def _inject_canned_actions(items):
             action._prepare_deps(items)
             added_actions[triggered_item_id] = action
 
-    return items + list(added_actions.values())
+    items.update({item.id: item for item in added_actions.values()})
+    return items
 
 
 def _inject_concurrency_blockers(items):
     """
-    Looks for items with BLOCK_CONCURRENT set to True and inserts
+    Looks for items with BLOCK_CONCURRENT set and inserts daisy-chain
     dependencies to force a sequential apply.
     """
     # find every item type that cannot be applied in parallel
     item_types = set()
-    for item in items:
-        item._concurrency_deps = []
+    for item in items.values():
+        item._concurrency_deps = []  # used for DOT (graphviz) output only
         if (
             not isinstance(item, DummyItem) and
             item.BLOCK_CONCURRENT
         ):
             item_types.add(item.__class__)
 
-    # daisy-chain all items of the blocking type and all items of the
-    # blocked types while respecting existing dependencies between them
+    # Now that we have collected all types with BLOCK_CONCURRENT,
+    # we must group them together when they overlap. E.g.:
+    #
+    #     Type1.BLOCK_CONCURRENT = ["type1", "type2"]
+    #     Type2.BLOCK_CONCURRENT = ["type2, "type3"]
+    #     Type4.BLOCK_CONCURRENT = ["type4"]
+    #
+    # becomes
+    #
+    #     ["type1", "type2", "type3"]
+    #     ["type4"]
+    #
+    # because the first two types overlap in blocking type2. This is
+    # necessary because existing dependencies from type3 to type1 need
+    # to be taken into account when generating the daisy-chains
+    # connecting the three types. If we processed blockers for Type1 and
+    # Type2 independently, we might end up with two very different
+    # chains for Type2, which may cause circular dependencies.
+
+    chain_groups = []
     for item_type in item_types:
-        blocked_types = item_type.BLOCK_CONCURRENT + [item_type.ITEM_TYPE_NAME]
-        type_items = _find_items_of_types(
+        block_concurrent = list(item_type.BLOCK_CONCURRENT) + [item_type.ITEM_TYPE_NAME]
+        found = False
+        for blocked_types in chain_groups:
+            for blocked_type in block_concurrent:
+                if blocked_type in blocked_types:
+                    blocked_types.extend(block_concurrent)
+                    found = True
+                    break
+        if not found:
+            chain_groups.append(block_concurrent)
+
+    # daisy-chain all items of the chain group while respecting existing
+    # dependencies between them
+    for blocked_types in chain_groups:
+        blocked_types = set(blocked_types)
+        type_items = list(_find_items_of_types(
             blocked_types,
             items,
-        )
+        ))
         processed_items = []
         for item in type_items:
             # disregard deps to items of other types
@@ -291,6 +332,9 @@ def _inject_concurrency_blockers(items):
                     item._flattened_deps.append(previous_item.id)
             previous_item = item
             processed_items.append(item)
+            # Now remove all deps on the processed item. This frees up
+            # items depending *only* on the processed item to be
+            # eligible for the next iteration of this loop.
             for other_item in type_items:
                 try:
                     other_item.__deps.remove(item.id)
@@ -305,14 +349,13 @@ def _inject_tag_items(items):
     item in the list. Returns the appended list.
     """
     tag_items = {}
-    items = list(items)
-    for item in items:
+    for item in items.values():
         for tag in item.tags:
             if tag not in tag_items:
                 tag_items[tag] = TagItem(tag)
             tag_items[tag]._deps.append(item.id)
-
-    return list(tag_items.values()) + items
+    items.update({item.id: item for item in tag_items.values()})
+    return items
 
 
 def _inject_type_items(items):
@@ -320,10 +363,8 @@ def _inject_type_items(items):
     Takes a list of items and adds dummy items depending on each type of
     item in the list. Returns the appended list.
     """
-    # first, find all types of items and add dummy deps
     type_items = {}
-    items = list(items)
-    for item in items:
+    for item in items.values():
         # create dummy items that depend on each item of their type
         item_type = item.id.split(":")[0]
         if item_type not in type_items:
@@ -335,7 +376,9 @@ def _inject_type_items(items):
             item_type = dep.split(":")[0]
             if item_type not in type_items:
                 type_items[item_type] = TypeItem(item_type)
-    return list(type_items.values()) + items
+
+    items.update({item.id: item for item in type_items.values()})
+    return items
 
 
 def _inject_reverse_dependencies(items):
@@ -348,22 +391,22 @@ def _inject_reverse_dependencies(items):
             item._deps.append(dep)
             item._reverse_deps.append(dep)
 
-    for item in items:
+    for item in items.values():
         item._reverse_deps = []
 
-    for item in items:
+    for item in items.values():
         for depending_item_id in item.needed_by:
             # bundle items
             if depending_item_id.startswith("bundle:"):
                 depending_bundle_name = depending_item_id.split(":")[1]
-                for depending_item in items:
+                for depending_item in items.values():
                     if depending_item.bundle.name == depending_bundle_name:
                         add_dep(depending_item, item.id)
 
             # tag items
             if depending_item_id.startswith("tag:"):
                 tag_name = depending_item_id.split(":")[1]
-                for depending_item in items:
+                for depending_item in items.values():
                     if tag_name in depending_item.tags:
                         add_dep(depending_item, item.id)
 
@@ -375,7 +418,17 @@ def _inject_reverse_dependencies(items):
 
             # single items
             else:
-                depending_item = find_item(depending_item_id, items)
+                try:
+                    depending_item = items[depending_item_id]
+                except KeyError:
+                    raise ItemDependencyError(_(
+                        "'{item}' in bundle '{bundle}' has a reverse dependency (needed_by)"
+                        "on '{dep}', which doesn't exist"
+                    ).format(
+                        item=item.id,
+                        bundle=item.bundle.name,
+                        dep=depending_item_id,
+                    ))
                 add_dep(depending_item, item.id)
     return items
 
@@ -385,17 +438,17 @@ def _inject_reverse_triggers(items):
     Looks for 'triggered_by' and 'precedes' attributes and turns them
     into standard triggers (defined on the opposing end).
     """
-    for item in items:
+    for item in items.values():
         for triggering_item_id in item.triggered_by:
-            triggering_item = find_item(triggering_item_id, items)
+            triggering_item = items[triggering_item_id]
             if triggering_item.id.startswith("bundle:"):  # bundle items
                 bundle_name = triggering_item.id.split(":")[1]
-                for actual_triggering_item in items:
+                for actual_triggering_item in items.values():
                     if triggering_item.bundle.name == bundle_name:
                         actual_triggering_item.triggers.append(item.id)
             elif triggering_item.id.startswith("tag:"):  # tag items
                 tag_name = triggering_item.id.split(":")[1]
-                for actual_triggering_item in items:
+                for actual_triggering_item in items.values():
                     if tag_name in triggering_item.tags:
                         actual_triggering_item.triggers.append(item.id)
             elif triggering_item.id.endswith(":"):  # type items
@@ -405,15 +458,15 @@ def _inject_reverse_triggers(items):
             else:
                 triggering_item.triggers.append(item.id)
         for preceded_item_id in item.precedes:
-            preceded_item = find_item(preceded_item_id, items)
+            preceded_item = items[preceded_item_id]
             if preceded_item.id.startswith("bundle:"):  # bundle items
                 bundle_name = preceded_item.id.split(":")[1]
-                for actual_preceded_item in items:
+                for actual_preceded_item in items.values():
                     if actual_preceded_item.bundle.name == bundle_name:
                         actual_preceded_item.preceded_by.append(item.id)
             elif preceded_item.id.startswith("tag:"):  # tag items
                 tag_name = preceded_item.id.split(":")[1]
-                for actual_preceded_item in items:
+                for actual_preceded_item in items.values():
                     if tag_name in actual_preceded_item.tags:
                         actual_preceded_item.preceded_by.append(item.id)
             elif preceded_item.id.endswith(":"):  # type items
@@ -430,11 +483,11 @@ def _inject_trigger_dependencies(items):
     Injects dependencies from all triggered items to their triggering
     items.
     """
-    for item in items:
+    for item in items.values():
         for triggered_item_id in item.triggers:
             try:
-                triggered_item = find_item(triggered_item_id, items)
-            except NoSuchItem:
+                triggered_item = items[triggered_item_id]
+            except KeyError:
                 raise BundleError(_(
                     "unable to find definition of '{item1}' triggered "
                     "by '{item2}' in bundle '{bundle}'"
@@ -463,7 +516,7 @@ def _inject_preceded_by_dependencies(items):
     Injects dependencies from all triggering items to their
     preceded_by items and attaches triggering items to preceding items.
     """
-    for item in items:
+    for item in items.values():
         if item.preceded_by and item.triggered:
             raise BundleError(_(
                 "triggered item '{item}' in bundle '{bundle}' must not use "
@@ -474,8 +527,8 @@ def _inject_preceded_by_dependencies(items):
             ))
         for triggered_item_id in item.preceded_by:
             try:
-                triggered_item = find_item(triggered_item_id, items)
-            except NoSuchItem:
+                triggered_item = items[triggered_item_id]
+            except KeyError:
                 raise BundleError(_(
                     "unable to find definition of '{item1}' preceding "
                     "'{item2}' in bundle '{bundle}'"
@@ -504,12 +557,14 @@ def prepare_dependencies(items):
     """
     Performs all dependency preprocessing on a list of items.
     """
-    items = list(items)
-
     for item in items:
         item._check_bundle_collisions(items)
+        item._check_loopback_dependency()
         item._prepare_deps(items)
 
+    # transform items into a dict to prevent repeated item.id lookups
+    items = {item.id: item for item in items}
+
     items = _inject_bundle_items(items)
     items = _inject_tag_items(items)
     items = _inject_type_items(items)
@@ -521,11 +576,11 @@ def prepare_dependencies(items):
     items = _flatten_dependencies(items)
     items = _inject_concurrency_blockers(items)
 
-    for item in items:
+    for item in items.values():
         if not isinstance(item, DummyItem):
             item._check_redundant_dependencies()
 
-    return items
+    return list(items.values())
 
 
 def remove_dep_from_items(items, dep):
@@ -587,11 +642,11 @@ def split_items_without_deps(items):
     Takes a list of items and extracts the ones that don't have any
     dependencies. The extracted deps are returned as a list.
     """
-    items = list(items)  # make sure we're not returning a generator
+    remaining_items = []
     removed_items = []
     for item in items:
-        if not item._deps:
+        if item._deps:
+            remaining_items.append(item)
+        else:
             removed_items.append(item)
-    for item in removed_items:
-        items.remove(item)
-    return (items, removed_items)
+    return (remaining_items, removed_items)
