@@ -6,17 +6,20 @@ Repository.item_classes loads them as files.
 from __future__ import unicode_literals
 from copy import copy
 from datetime import datetime
+from inspect import cleandoc
 from os.path import join
+from textwrap import TextWrapper
 
-from bundlewrap.exceptions import BundleError, FaultUnavailable
+from bundlewrap.exceptions import BundleError, ItemDependencyError, FaultUnavailable
 from bundlewrap.utils import cached_property
 from bundlewrap.utils.statedict import diff_keys, diff_value, hash_statedict, validate_statedict
 from bundlewrap.utils.text import force_text, mark_for_translation as _
-from bundlewrap.utils.text import blue, bold, wrap_question
+from bundlewrap.utils.text import blue, bold, italic, wrap_question
 from bundlewrap.utils.ui import io
 
 BUILTIN_ITEM_ATTRIBUTES = {
     'cascade_skip': None,
+    'comment': None,
     'needed_by': [],
     'needs': [],
     'preceded_by': [],
@@ -27,8 +30,24 @@ BUILTIN_ITEM_ATTRIBUTES = {
     'triggered_by': [],
     'triggers': [],
     'unless': "",
+    'when_creating': {},
 }
 
+wrapper = TextWrapper(
+    break_long_words=False,
+    break_on_hyphens=False,
+    expand_tabs=False,
+    replace_whitespace=False,
+)
+
+
+def format_comment(comment):
+    result = "\n\n"
+    for line in wrapper.wrap(cleandoc(comment)):
+        for inlineline in line.split("\n"):
+            result += "# {}\n".format(italic(inlineline))
+    return result
+
 
 class ItemStatus(object):
     """
@@ -57,7 +76,6 @@ class Item(object):
     """
     A single piece of configuration (e.g. a file, a package, a service).
     """
-    BINARY_ATTRIBUTES = []
     BLOCK_CONCURRENT = []
     BUNDLE_ATTRIBUTE_NAME = None
     ITEM_ATTRIBUTES = {}
@@ -68,6 +86,7 @@ class Item(object):
     STATUS_FAILED = 3
     STATUS_SKIPPED = 4
     STATUS_ACTION_SUCCEEDED = 5
+    WHEN_CREATING_ATTRIBUTES = {}
 
     def __init__(
         self,
@@ -85,6 +104,7 @@ class Item(object):
         self.item_data_dir = join(bundle.bundle_data_dir, self.BUNDLE_ATTRIBUTE_NAME)
         self.name = name
         self.node = bundle.node
+        self.when_creating = {}
         self._faults_missing_for_attributes = set()
         self._precedes_items = []
 
@@ -101,15 +121,13 @@ class Item(object):
         except FaultUnavailable:
             self._faults_missing_for_attributes.add(_("unknown"))
 
-        for attribute_name, attribute_default in \
-                BUILTIN_ITEM_ATTRIBUTES.items():
+        for attribute_name, attribute_default in BUILTIN_ITEM_ATTRIBUTES.items():
             setattr(self, attribute_name, force_text(attributes.get(
                 attribute_name,
                 copy(attribute_default),
             )))
 
-        for attribute_name, attribute_default in \
-                self.ITEM_ATTRIBUTES.items():
+        for attribute_name, attribute_default in self.ITEM_ATTRIBUTES.items():
             if attribute_name not in BUILTIN_ITEM_ATTRIBUTES:
                 try:
                     self.attributes[attribute_name] = force_text(attributes.get(
@@ -119,6 +137,10 @@ class Item(object):
                 except FaultUnavailable:
                     self._faults_missing_for_attributes.add(attribute_name)
 
+        for attribute_name, attribute_default in self.WHEN_CREATING_ATTRIBUTES.items():
+            self.when_creating[attribute_name] = \
+                attributes.get('when_creating', {}).get(attribute_name, attribute_default)
+
         if self.cascade_skip is None:
             self.cascade_skip = not (self.unless or self.triggered)
 
@@ -152,6 +174,19 @@ class Item(object):
                     bundle2=self.bundle.name,
                 ))
 
+    def _check_loopback_dependency(self):
+        """
+        Alerts the user if they have an item depend on itself.
+        """
+        if self.id in self.needs or self.id in self.needed_by:
+            raise ItemDependencyError(_(
+                "'{item}' in bundle '{bundle}' on node '{node}' cannot depend on itself"
+            ).format(
+                item=self.id,
+                bundle=self.bundle.name,
+                node=self.node.name,
+            ))
+
     def _check_redundant_dependencies(self):
         """
         Alerts the user if they have defined a redundant dependency
@@ -160,7 +195,7 @@ class Item(object):
         """
         for dep in self._deps:
             if self._deps.count(dep) > 1:
-                raise BundleError(_(
+                raise ItemDependencyError(_(
                     "redundant dependency of {item1} in bundle '{bundle}' on {item2}"
                 ).format(
                     bundle=self.bundle.name,
@@ -286,13 +321,25 @@ class Item(object):
             ),
         )
         if invalid_attributes:
-            raise BundleError(
-                _("invalid attribute(s) for '{item}' in bundle '{bundle}': {attrs}").format(
-                    item=item_id,
-                    bundle=bundle.name,
-                    attrs=", ".join(invalid_attributes),
-                )
-            )
+            raise BundleError(_(
+                "invalid attribute(s) for '{item}' in bundle '{bundle}': {attrs}"
+            ).format(
+                item=item_id,
+                bundle=bundle.name,
+                attrs=", ".join(invalid_attributes),
+            ))
+
+        invalid_attributes = set(attributes.get('when_creating', {}).keys()).difference(
+            set(cls.WHEN_CREATING_ATTRIBUTES.keys())
+        )
+        if invalid_attributes:
+            raise BundleError(_(
+                "invalid when_creating attribute(s) for '{item}' in bundle '{bundle}': {attrs}"
+            ).format(
+                item=item_id,
+                bundle=bundle.name,
+                attrs=", ".join(invalid_attributes),
+            ))
 
     @classmethod
     def _validate_name(cls, bundle, name):
@@ -426,6 +473,8 @@ class Item(object):
                         keys_to_fix,
                     )
                     question_text = self.ask(cdict, sdict, keys_to_fix)
+                if self.comment:
+                    question_text += format_comment(self.comment)
                 question = wrap_question(
                     self.id,
                     question_text,
@@ -498,8 +547,8 @@ class Item(object):
     def cdict(self):
         """
         Return a statedict that describes the target state of this item
-        as configured in the repo. An empty dict means that the item
-        should not exist.
+        as configured in the repo. Returning `None` instead means that
+        the item should not exist.
 
         MAY be overridden by subclasses.
         """
@@ -577,6 +626,9 @@ class Item(object):
             return self.name
         return "{}:{}".format(self.ITEM_TYPE_NAME, self.name)
 
+    def verify(self):
+        return self.cached_unless_result, self.cached_status
+
     def display_dicts(self, cdict, sdict, keys):
         """
         Given cdict and sdict as implemented above, modify them to
@@ -609,8 +661,8 @@ class Item(object):
     def sdict(self):
         """
         Return a statedict that describes the actual state of this item
-        on the node. An empty dict means that the item does not exist
-        on the node.
+        on the node. Returning `None` instead means that the item does
+        not exist on the node.
 
         For the item to validate as correct, the values for all keys in
         self.cdict() have to match this statedict.
