@@ -1,3 +1,6 @@
+# -*- coding: utf-8 -*-
+from __future__ import unicode_literals
+
 from datetime import datetime
 from getpass import getuser
 import json
@@ -38,7 +41,7 @@ class NodeLock(object):
                 with io.job(_("  {node}  checking hard lock status...").format(node=self.node.name)):
                     result = self.node.run("mkdir " + quote(HARD_LOCK_PATH), may_fail=True)
                     if result.return_code != 0:
-                        self.node.download(HARD_LOCK_FILE, local_path, ignore_failure=True)
+                        self.node.download(HARD_LOCK_FILE, local_path)
                         with open(local_path, 'r') as f:
                             try:
                                 info = json.loads(f.read())
@@ -159,6 +162,8 @@ def softlock_add(node, lock_id, comment="", expiry="8h", item_selectors=None):
         node.run("mkdir -p " + quote(SOFT_LOCK_PATH))
         node.upload(local_path, SOFT_LOCK_FILE.format(id=lock_id), mode='0644')
 
+    node.repo.hooks.lock_add(node.repo, node, lock_id, item_selectors, expiry_timestamp, comment)
+
     return lock_id
 
 
@@ -196,3 +201,4 @@ def softlock_remove(node, lock_id):
         node=node.name,
     ))
     node.run("rm {}".format(SOFT_LOCK_FILE.format(id=lock_id)))
+    node.repo.hooks.lock_remove(node.repo, node, lock_id)
