@@ -2,6 +2,7 @@
 from __future__ import unicode_literals
 
 from datetime import datetime, timedelta
+from hashlib import md5
 from os import environ
 from sys import exit
 from threading import Lock
@@ -16,7 +17,7 @@ from .deps import (
 from .exceptions import (
     DontCache,
     FaultUnavailable,
-    ItemDependencyError,
+    ItemDependencyLoop,
     NodeLockedException,
     NoSuchBundle,
     RepositoryError,
@@ -26,10 +27,11 @@ from .itemqueue import ItemQueue, ItemTestQueue
 from .items import Item
 from .lock import NodeLock
 from .metadata import check_for_unsolvable_metadata_key_conflicts, hash_metadata
-from .utils import cached_property, graph_for_items, names
+from .utils import cached_property, names
 from .utils.statedict import hash_statedict
 from .utils.text import blue, bold, cyan, green, red, validate_name, yellow
 from .utils.text import force_text, mark_for_translation as _
+from .utils.time import format_duration
 from .utils.ui import io
 
 
@@ -131,6 +133,7 @@ def apply_items(
     def next_task():
         item, skipped_items = item_queue.pop()
         for skipped_item in skipped_items:
+            io.progress_advance()
             handle_apply_result(
                 node,
                 skipped_item,
@@ -197,6 +200,7 @@ def apply_items(
 
         handle_apply_result(node, item, status_code, interactive, changes=changes)
         if not isinstance(item, DummyItem):
+            io.progress_advance()
             results.append((item.id, status_code, duration))
 
     worker_pool = WorkerPool(
@@ -211,17 +215,7 @@ def apply_items(
     # we have no items without deps left and none are processing
     # there must be a loop
     if item_queue.items_with_deps:
-        io.debug(_(
-            "There was a dependency problem. Look at the debug.svg generated "
-            "by the following command and try to find a loop:\n"
-            "printf '{}' | dot -Tsvg -odebug.svg"
-        ).format("\\n".join(graph_for_items(node.name, item_queue.items_with_deps))))
-
-        raise ItemDependencyError(
-            _("bad dependencies between these items: {}").format(
-                ", ".join([i.id for i in item_queue.items_with_deps]),
-            )
-        )
+        raise ItemDependencyLoop(item_queue.items_with_deps)
 
     return results
 
@@ -332,6 +326,7 @@ class Node(object):
         'rhel',
         'centos',
         'fedora',
+        'oraclelinux',
     )
 
     OS_FAMILY_LINUX = (
@@ -340,34 +335,42 @@ class Node(object):
         'opensuse',
         'gentoo',
         'linux',
-        'oraclelinux',
     ) + \
         OS_FAMILY_DEBIAN + \
         OS_FAMILY_REDHAT
 
     OS_KNOWN = OS_FAMILY_BSD + OS_FAMILY_LINUX
 
-    def __init__(self, name, infodict=None):
-        if infodict is None:
-            infodict = {}
+    def __init__(self, name, attributes=None, transport='ssh', transport_options=None):
+        if attributes is None:
+            attributes = {}
+        if transport_options is None:
+            transport_options = {}
+
+        if transport == 'ssh':
+            transport_options.setdefault(
+                'add_host_keys',
+                environ.get('BW_ADD_HOST_KEYS', False) == "1",
+            )
 
         if not validate_name(name):
             raise RepositoryError(_("'{}' is not a valid node name").format(name))
 
-        self.name = name
-        self._bundles = infodict.get('bundles', [])
+        self._bundles = attributes.get('bundles', [])
         self._compiling_metadata = Lock()
         self._dynamic_group_lock = Lock()
         self._dynamic_groups_resolved = False  # None means we're currently doing it
         self._metadata_so_far = {}
-        self._node_metadata = infodict.get('metadata', {})
+        self._node_metadata = attributes.get('metadata', {})
         self._ssh_conn_established = False
         self._ssh_first_conn_lock = Lock()
-        self.add_ssh_host_keys = False
-        self.hostname = infodict.get('hostname', self.name)
+        self.hostname = attributes.get('hostname', name)
+        self.name = name
+        self.transport = transport
+        self.transport_options = transport_options
 
         for attr in GROUP_ATTR_DEFAULTS:
-            setattr(self, "_{}".format(attr), infodict.get(attr))
+            setattr(self, "_{}".format(attr), attributes.get(attr))
 
     def __lt__(self, other):
         return self.name < other.name
@@ -449,19 +452,27 @@ class Node(object):
 
         # we have to add parent groups at the very end, since we might
         # have added or removed subgroups thru .members_add/remove
-        for group in list(_groups):
-            for parent_group in group.parent_groups:
-                if cache_result:
-                    with self._dynamic_group_lock:
-                        self._dynamic_groups_resolved = None
-                        if (
-                            not parent_group.members_remove or
-                            not parent_group.members_remove(self)
-                        ):
-                            _groups.add(parent_group)
-                        self._dynamic_groups_resolved = True
-                else:
-                    _groups.add(parent_group)
+        while True:
+            # Since we're only looking at *immediate* parent groups,
+            # we have to keep doing this until we stop adding parent
+            # groups.
+            _original_groups = _groups.copy()
+            for group in list(_groups):
+                for parent_group in group.immediate_parent_groups:
+                    if cache_result:
+                        with self._dynamic_group_lock:
+                            self._dynamic_groups_resolved = None
+                            if (
+                                not parent_group.members_remove or
+                                not parent_group.members_remove(self)
+                            ):
+                                _groups.add(parent_group)
+                            self._dynamic_groups_resolved = True
+                    else:
+                        _groups.add(parent_group)
+            if _groups == _original_groups:
+                # we didn't add any new parent groups, so we can stop
+                break
 
         if cache_result:
             return sorted(_groups)
@@ -518,6 +529,10 @@ class Node(object):
                 for item in bundle.items:
                     yield item
 
+    @cached_property
+    def magic_number(self):
+        return int(md5(self.name.encode('UTF-8')).hexdigest(), 16)
+
     @property
     def _static_items(self):
         for bundle in self.bundles:
@@ -529,21 +544,36 @@ class Node(object):
         autoskip_selector="",
         interactive=False,
         force=False,
+        skip_list=tuple(),
         workers=4,
         profiling=False,
     ):
         if not list(self.items):
-            io.stdout(_("{x} {node}  has no items").format(node=bold(self.name), x=yellow("!")))
+            io.stdout(_("{x} {node}  has no items").format(
+                node=bold(self.name),
+                x=yellow("»"),
+            ))
             return None
 
         if self.covered_by_autoskip_selector(autoskip_selector):
-            io.debug(_("skipping {}, matches autoskip selector").format(self.name))
+            io.stdout(_("{x} {node}  skipped by --skip").format(
+                node=bold(self.name),
+                x=yellow("»"),
+            ))
+            return None
+
+        if self.name in skip_list:
+            io.stdout(_("{x} {node}  skipped by --resume-file").format(
+                node=bold(self.name),
+                x=yellow("»"),
+            ))
             return None
 
         start = datetime.now()
 
-        io.stdout(_("{x} {node} run started at {time}").format(
+        io.stdout(_("{x} {node}  {started} at {time}").format(
             node=bold(self.name),
+            started=bold(_("started")),
             time=start.strftime("%Y-%m-%d %H:%M:%S"),
             x=blue("i"),
         ))
@@ -566,7 +596,10 @@ class Node(object):
                 )
         except NodeLockedException as e:
             if not interactive:
-                io.stderr(_("{x} {node} already locked by {user} at {date} ({duration} ago, `bw apply -f` to override)").format(
+                io.stderr(_(
+                    "{x} {node} already locked by {user} at {date} ({duration} ago, "
+                    "`bw apply -f` to override)"
+                ).format(
                     date=bold(e.args[0]['date']),
                     duration=e.args[0]['duration'],
                     node=bold(self.name),
@@ -578,14 +611,11 @@ class Node(object):
         result.start = start
         result.end = datetime.now()
 
-        io.stdout(_("{x} {node} run completed after {time}s").format(
-            node=bold(self.name),
-            time=(result.end - start).total_seconds(),
-            x=blue("i"),
-        ))
-        io.stdout(_("{x} {node} stats: {stats}").format(
+        io.stdout(_("{x} {node}  {completed} after {time}  ({stats})").format(
+            completed=bold(_("completed")),
             node=bold(self.name),
             stats=format_node_result(result),
+            time=format_duration(result.end - start),
             x=blue("i"),
         ))
 
@@ -599,12 +629,12 @@ class Node(object):
 
         return result
 
-    def download(self, remote_path, local_path, ignore_failure=False):
+    def download(self, remote_path, local_path):
         return operations.download(
             self.hostname,
             remote_path,
             local_path,
-            add_host_keys=True if environ.get('BW_ADD_HOST_KEYS', False) == "1" else False,
+            add_host_keys=self.transport_options['add_host_keys'],
             wrapper_inner=self.cmd_wrapper_inner,
             wrapper_outer=self.cmd_wrapper_outer,
         )
@@ -654,7 +684,7 @@ class Node(object):
         """
         return self.repo._metadata_for_node(self.name, partial=True)
 
-    def run(self, command, may_fail=False, log_output=False):
+    def run(self, command, data_stdin=None, may_fail=False, log_output=False):
         if log_output:
             def log_function(msg):
                 io.stdout("{x} {node}  {msg}".format(
@@ -665,7 +695,7 @@ class Node(object):
         else:
             log_function = None
 
-        add_host_keys = True if environ.get('BW_ADD_HOST_KEYS', False) == "1" else False
+        add_host_keys = self.transport_options['add_host_keys']
 
         if not self._ssh_conn_established:
             # Sometimes we're opening SSH connections to a node too fast
@@ -690,6 +720,7 @@ class Node(object):
             self.hostname,
             command,
             add_host_keys=add_host_keys,
+            data_stdin=data_stdin,
             ignore_failure=may_fail,
             log_function=log_function,
             wrapper_inner=self.cmd_wrapper_inner,
@@ -715,7 +746,7 @@ class Node(object):
             self.hostname,
             local_path,
             remote_path,
-            add_host_keys=True if environ.get('BW_ADD_HOST_KEYS', False) == "1" else False,
+            add_host_keys=self.transport_options['add_host_keys'],
             group=group,
             mode=mode,
             owner=owner,
@@ -726,6 +757,8 @@ class Node(object):
     def verify(self, show_all=False, workers=4):
         bad = 0
         good = 0
+        start = datetime.now()
+
         if not self.items:
             io.stdout(_("{x} {node}  has no items").format(node=bold(self.name), x=yellow("!")))
         else:
@@ -739,7 +772,7 @@ class Node(object):
                 else:
                     bad += 1
 
-        return {'good': good, 'bad': bad}
+        return {'good': good, 'bad': bad, 'duration': datetime.now() - start}
 
 
 def build_attr_property(attr, default):
@@ -804,6 +837,9 @@ def test_items(node, ignore_missing_faults=False, workers=1):
 
     def handle_result(task_id, return_value, duration):
         node_name, bundle_name, item_id = task_id.split(":", 2)
+        if item_id.count(":") < 2:
+            # don't count canned actions
+            io.progress_advance()
         io.stdout("{x} {node}  {bundle}  {item}".format(
             bundle=bold(bundle_name),
             item=item_id,
@@ -812,6 +848,7 @@ def test_items(node, ignore_missing_faults=False, workers=1):
         ))
 
     def handle_exception(task_id, exception, traceback):
+        io.progress_advance()
         node_name, bundle_name, item_id = task_id.split(":", 2)
         if ignore_missing_faults and isinstance(exception, FaultUnavailable):
             io.stderr(_("{x} {node}  {bundle}  {item}  ({msg})").format(
@@ -843,17 +880,7 @@ def test_items(node, ignore_missing_faults=False, workers=1):
     worker_pool.run()
 
     if item_queue.items_with_deps:
-        io.stderr(_(
-            "There was a dependency problem. Look at the debug.svg generated "
-            "by the following command and try to find a loop:\n"
-            "printf '{}' | dot -Tsvg -odebug.svg"
-        ).format("\\n".join(graph_for_items(node.name, item_queue.items_with_deps))))
-
-        raise ItemDependencyError(
-            _("bad dependencies between these items: {}").format(
-                ", ".join([i.id for i in item_queue.items_with_deps]),
-            )
-        )
+        raise ItemDependencyLoop(item_queue.items_with_deps)
 
 
 def verify_items(node, show_all=False, workers=1):
@@ -864,6 +891,8 @@ def verify_items(node, show_all=False, workers=1):
             not item.triggered
         ):
             items.append(item)
+        elif not isinstance(item, DummyItem):
+            io.progress_advance()
 
     def tasks_available():
         return bool(items)
@@ -878,6 +907,7 @@ def verify_items(node, show_all=False, workers=1):
                 if item.error_on_missing_fault:
                     item._raise_for_faults()
                 else:
+                    io.progress_advance()
                     io.stdout(_("{x} {node}  {bundle}  {item}  ({msg})").format(
                         bundle=bold(item.bundle.name),
                         item=item.id,
@@ -888,12 +918,14 @@ def verify_items(node, show_all=False, workers=1):
             else:
                 return {
                     'task_id': node.name + ":" + item.bundle.name + ":" + item.id,
-                    'target': item.get_status,
+                    'target': item.verify,
                 }
 
-    def handle_result(task_id, item_status, duration):
+    def handle_result(task_id, return_value, duration):
+        io.progress_advance()
+        unless_result, item_status = return_value
         node_name, bundle_name, item_id = task_id.split(":", 2)
-        if not item_status.correct:
+        if not unless_result and not item_status.correct:
             if item_status.must_be_created:
                 changes_text = _("create")
             elif item_status.must_be_deleted:
