@@ -2,6 +2,7 @@
 from __future__ import unicode_literals
 
 from imp import load_source
+from inspect import isabstract
 from os import listdir, mkdir
 from os.path import isdir, isfile, join
 from threading import Lock
@@ -38,18 +39,21 @@ FILENAME_NODES = "nodes.py"
 FILENAME_REQUIREMENTS = "requirements.txt"
 
 HOOK_EVENTS = (
-    'action_run_start',
     'action_run_end',
-    'apply_start',
+    'action_run_start',
     'apply_end',
-    'item_apply_start',
+    'apply_start',
     'item_apply_end',
-    'node_apply_start',
+    'item_apply_start',
+    'lock_add',
+    'lock_remove',
+    'lock_show',
     'node_apply_end',
-    'node_run_start',
+    'node_apply_start',
     'node_run_end',
-    'run_start',
+    'node_run_start',
     'run_end',
+    'run_start',
     'test',
     'test_node',
 )
@@ -191,7 +195,7 @@ def items_from_path(path):
             if obj == items.Item or name.startswith("_"):
                 continue
             try:
-                if issubclass(obj, items.Item):
+                if issubclass(obj, items.Item) and not isabstract(obj):
                     yield obj
             except TypeError:
                 pass
@@ -463,6 +467,8 @@ class Repository(object):
         self._node_metadata_partial.keys().
         """
         iterations = {}
+        # these processors have indicated that they do not need to be run again
+        blacklisted_metaprocs = set()
         while (
             not iterations or max(iterations.values()) <= META_PROC_MAX_ITER
         ) and not QUIT_EVENT.is_set():
@@ -500,6 +506,8 @@ class Repository(object):
                 node = self.get_node(node_name)
                 with io.job(_("  {node}  running metadata processors...").format(node=node.name)):
                     for metadata_processor_name, metadata_processor in node.metadata_processors:
+                        if (node_name, metadata_processor_name) in blacklisted_metaprocs:
+                            continue
                         iterations.setdefault((node.name, metadata_processor_name), 1)
                         io.debug(_(
                             "running metadata processor {metaproc} for node {node}, "
@@ -509,14 +517,36 @@ class Repository(object):
                             node=node.name,
                             i=iterations[(node.name, metadata_processor_name)],
                         ))
-                        processed = metadata_processor(
-                            deepcopy_metadata(self._node_metadata_partial[node.name]),
-                        )
+                        try:
+                            processed = metadata_processor(
+                                deepcopy_metadata(self._node_metadata_partial[node.name]),
+                            )
+                        except Exception as exc:
+                            io.stderr(_(
+                                "{x} Exception while executing metadata processor "
+                                "{metaproc} for node {node}:"
+                            ).format(
+                                x=red("!!!"),
+                                metaproc=metadata_processor_name,
+                                node=node.name,
+                            ))
+                            raise exc
                         iterations[(node.name, metadata_processor_name)] += 1
+                        if isinstance(processed, tuple) and len(processed) == 2:
+                            if processed[1] is True:
+                                io.debug(_(
+                                    "metadata processor {metaproc} for node {node} "
+                                    "has indicated that it need not be run again"
+                                ).format(
+                                    metaproc=metadata_processor_name,
+                                    node=node.name,
+                                ))
+                                blacklisted_metaprocs.add((node_name, metadata_processor_name))
+                            processed = processed[0]
                         if not isinstance(processed, dict):
                             raise ValueError(_(
                                 "metadata processor {metaproc} for node {node} did not return "
-                                "a dictionary"
+                                "a dictionary or tuple of (dict, bool)"
                             ).format(
                                 metaproc=metadata_processor_name,
                                 node=node.name,
