@@ -15,6 +15,10 @@ ANSI_ESCAPE = re.compile(r'\x1b[^m]*m')
 VALID_NAME_CHARS = digits + ascii_letters + "-_.+"
 
 
+def ansi_clean(input_string):
+    return ANSI_ESCAPE.sub("", force_text(input_string))
+
+
 def ansi_wrapper(colorizer):
     if environ.get("BW_COLORS", "1") != "0":
         return colorizer
@@ -42,6 +46,11 @@ def inverse(text):
     return "\033[0m\033[7m{}\033[0m".format(text)
 
 
+@ansi_wrapper
+def italic(text):
+    return "\033[3m{}\033[0m".format(text)
+
+
 @ansi_wrapper
 def green(text):
     return "\033[32m{}\033[0m".format(text)
@@ -57,6 +66,27 @@ def yellow(text):
     return "\033[33m{}\033[0m".format(text)
 
 
+def green_unless_zero(number):
+    if number == 0:
+        return "0"
+    else:
+        return green(str(number))
+
+
+def red_unless_zero(number):
+    if number == 0:
+        return "0"
+    else:
+        return red(str(number))
+
+
+def yellow_unless_zero(number):
+    if number == 0:
+        return "0"
+    else:
+        return yellow(str(number))
+
+
 def error_summary(errors):
     if not errors:
         return
